/*
 * AdabasApi.h --
 *
 *      This module contains procedures, that implement the functionalty
 *      of the adasql procedures (adalogon, adaopen, adasql, adafetch, ...).
 *      Despite the fact, that all procedures result Tcl results, it
 *      could be used by other extensions (perl, python, ...), too.
 *
 * Copyright (c) 1996-1997 Christian Krone. All rights reserved.
 * This program is free software; you can redistribute it and/or
 * modify it under the same terms as Tcl itself.
 * See also licence.terms
 *
 *  $Id: AdabasApi.c,v 1.12 1997/01/14 20:55:13 adabas Exp $
 */

/* ----------------- Import interface --------------------------- */
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <tcl.h>

#include "adabas.h"
#include "AdabasPort.h"
#include "AdabasRte.h"
#include "AdabasPacket.h"
#include "AdabasFormat.h"
#include "AdabasUtil.h"
#include "AdabasSend.h"
#include "AdabasApi.h"
#include "AdabasLong.h"


/* ----------------- Local constants ---------------------------- */
#define MAX_CURSORS  25

#define USER_RANGE   1
#define PASSWD_RANGE 2

#define CMD_RANGE    1

/* ----------------- Type definitions --------------------------- */

typedef enum {
  noSelect,
  massSelect,
  intoSelect,
  showSelect,
  explainSelect
} SelectKind;

/* ----------------- Local procedure headings ------------------- */
static int  selectInto _ANSI_ARGS_((Tcl_Interp *interp, CursorInfo *cursorVars,
				    char *selectInto));
static void invalidateResultTable _ANSI_ARGS_((CursorInfo *cursorVars));
static int executeWithParameter _ANSI_ARGS_((Tcl_Interp *interp,
					     CursorInfo  *cursorVars,
					     Tcl_DString *parsId));
static void examineSqlCommand _ANSI_ARGS_((Tcl_Interp *interp, char *cmdString,
					   int *withInfo, int *withResCount,
					   SelectKind *selectKind,
					   char **selectInto));
static char *getNextData _ANSI_ARGS_((Tcl_Interp *interp, char *data,
				      CursorInfo *cursorVars));
static int appendOneValue _ANSI_ARGS_((Tcl_Interp *interp, char *data,
				       Tcl_Obj *destObj, char *nullValue,
				       char *specialNull, field_info *currSi,
				       int *longColFound, int isElement));
static void getNullValue _ANSI_ARGS_((Tcl_Interp *interp, char **nullValue,
				       char **specialNull));
static int checkCursor _ANSI_ARGS_((Tcl_Interp *interp, CursorInfo *cursor));

/* ----------------- Local variables ---------------------------- */

/*
 * This shouldn't be here, instead there should be a list
 * of all cursors starting in the logonHandle.
 */

static Tcl_HashTable cursorHash;

/* ----------------- Exported procedures ------------------------- */

void
AdabasApiInit (interp)
     Tcl_Interp *interp;		/* Current interpreter. */
{
  char version[200];

  SET_VERSION (version);
  AdabasSetMsgString (interp, AdamsgVersion,     version);
  AdabasSetMsgString (interp, AdamsgNullvalue,   "");
  AdabasSetMsgString (interp, AdamsgSpecialnull, "***");

  /*
   * Initialize the only global variable, we have: the hash table of cursors.
   */

  Tcl_InitHashTable (&cursorHash, TCL_STRING_KEYS);
}

/* --------------------------------------------------------------- */

char *
getDataOfCursor (interp, cursorVars, flags)
     Tcl_Interp *interp;		/* Current interpreter. */
     CursorInfo *cursorVars;		/* Current cursor. */
     int         flags;			/* defined in AdabasPacket.h */
{
  char *data;

#if ADABAS_VERSION >= 62
  data = getData (&cursorVars->logonInfo->rcvPacket->segm,
		  &cursorVars->dataLength, &cursorVars->massRows, flags);

#else /* ADABAS_VERSION <= 61 */
  data = getData (&cursorVars->logonInfo->sqlPacket,
		  &cursorVars->dataLength, &cursorVars->massRows, flags);
#endif

  cursorVars->rowIx = 0;

  if (!data) {
    SetResult (interp, "Oops, no data part found???");
    return (char *) NULL;
  }
  return data;
}

/* ----------------- Local procedures --------------------------- */

CursorInfo *
AdabasOpen (interp, vars)
     Tcl_Interp *interp;		/* Current interpreter. */
     AdabasInfo *vars;			/* Current logon. */
{
  CursorInfo    *cursorVars;		/* the newly created cursorHandle. */
  Tcl_HashEntry *newEntry;		/* Entry with cursor name and infoPrt. */
  int            added;			/* Hash entry successfulle added? */
  int            cnt;			/* Count above cursor. */

  Tcl_ResetObjResult (interp);
  if (AdabasCheckLogon (interp, vars) == TCL_ERROR) {
    return (CursorInfo *) NULL;
  }
    
  if ((cursorVars = (CursorInfo *) ckalloc (sizeof (CursorInfo))) == NULL) {
    AppendResult (interp, "Allocation of cursor failed due to lack of memory");
    return (CursorInfo *) NULL;
  }
  cursorVars->logonInfo    = vars;
  cursorVars->cntParams    = 0;
  cursorVars->longColFound = 0;
  cursorVars->dataLength   = 0;
  cursorVars->dataEnd      = 0;
  cursorVars->shortInfos   = NULL;
  cursorVars->columnNames  = NULL;
  cursorVars->magic        = CURSOR_MAGIC;

  /*
   * Store the new cursorHandle into our hashtable with its name as key
   * Put the name of the cursor into the CursorInfo.
   */

  added = 0;
  cnt   = 0;
  while (!added) {
    sprintf (cursorVars->cursorName, "cursor%d", ++cnt);
    newEntry = Tcl_CreateHashEntry (&cursorHash, cursorVars->cursorName, &added);
  }
  Tcl_SetHashValue (newEntry, cursorVars);

  return cursorVars;
}

/* -------------------------------------------------------------- */

int
AdabasClose (interp, cursorVars)
     Tcl_Interp *interp;		/* Current interpreter. */
     CursorInfo *cursorVars;		/* Cursor to be freed. */
{
  Tcl_HashEntry *oldEntry;		/* hash table entry of cursorName. */

  if (checkCursor (interp, cursorVars) == TCL_ERROR) {
    return TCL_ERROR;
  }
  if (!(oldEntry = Tcl_FindHashEntry (&cursorHash, cursorVars->cursorName))) {
    SetResult (interp, "Oops, cursor not found in hash table???");
    return TCL_ERROR;
  }

  if (cursorVars->columnNames) {
    ckfree ((char *) cursorVars->columnNames);
  }
  if (cursorVars->shortInfos) {
    ckfree ((char *) cursorVars->shortInfos);
  }
  cursorVars->magic = 0;
  ckfree ((char *) cursorVars);

  /*
   * Delete the cursorHandle from our hashtable with its name as key.
   */

  Tcl_DeleteHashEntry (oldEntry);
  return TCL_OK;
}

/* -------------------------------------------------------------- */

int
AdabasSql (interp, cursorVars, messType, withParameter,
	   cmd, resultTable, sqlMode)
     Tcl_Interp *interp;		/* Current interpreter. */
     CursorInfo *cursorVars;		/* Current Cursor. */
     int         messType;		/* Message type (DBS, PARSE, EXECUTE) */
     int         withParameter;		/* Execute with parameter? */
     char       *cmd;			/* Sql command to execute. */
     char       *resultTable;		/* Name of resulttable (or NULL). */
     char       *sqlMode;		/* Wanted SQL mode (or NULL). */
{
  AdabasInfo *logonHandle;
  int         withInfo;
  int         withResCount;
  int         orgMessType;
  char       *selInto;
  long        resCnt;
  int         ret;
  int         oldSqlMode  = -1;
  int         errorOffset = 0;
  long        pos;
  SelectKind  selectKind;
  Tcl_DString cmdStr;
  Tcl_DString parsId;			/* ParsId returned, if only parsing. */
#if ADABAS_VERSION >= 62
  segment    *rcvSegment;
#else /* ADABAS_VERSION <= 61 */
  Tcl_RegExp  regExp;
  char       *start;
  char       *end;
#endif
  int         describeNeeded;
  Tcl_DString describe;
  char       *p;
  int         sqlModeInt = -1;

  Tcl_ResetObjResult (interp);
  if (checkCursor (interp, cursorVars) == TCL_ERROR) {
    return TCL_ERROR;
  }

  if (sqlMode && *sqlMode) {
    sqlModeInt = scanSqlMode (interp, sqlMode, 0);
#if ADABAS_VERSION >= 62
    if (sqlModeInt == SM_NIL) return TCL_ERROR;
#else /* ADABAS_VERSION <= 61 */
    if (sqlModeInt < 0) return TCL_ERROR;
#endif
  }

  cursorVars->longColFound = 0;
  cursorVars->dataLength   = 0;

  /*
   * Note, that adabas doesn't like Newlines in the command...
   */

  for (p = cmd; *p; p++) if (*p == '\n') *p = ' ';

  /*
   * Look at the command string and determine the kind of statement
   * (data manipulation, select, select into or the like)
   */

  examineSqlCommand (interp, cmd, &withInfo, &withResCount,
		     &selectKind, &selInto);

  if (selectKind == massSelect && resultTable && *resultTable) {
    Tcl_DStringInit   (&cmdStr);
    Tcl_DStringAppend (&cmdStr,  "DECLARE ",    -1);
    Tcl_DStringAppend (&cmdStr,   resultTable,  -1);
    Tcl_DStringAppend (&cmdStr, " CURSOR FOR ", -1);
    errorOffset = strlen (Tcl_DStringValue (&cmdStr));

    Tcl_DStringAppend (&cmdStr, cmd, -1);
    cmd = Tcl_DStringValue (&cmdStr);
  }

  logonHandle = cursorVars->logonInfo;

  if (sqlModeInt >= 0) {
    oldSqlMode = logonHandle->packetInfo.sqlMode;
    logonHandle->packetInfo.sqlMode = sqlModeInt;
  }

  orgMessType = logonHandle->packetInfo.messType;
  logonHandle->packetInfo.messType = messType;
  ret = AdabasSendCmdPacket (interp, logonHandle, withInfo, 1, cmd);
  logonHandle->packetInfo.messType = orgMessType;

#if ADABAS_VERSION >= 62
  rcvSegment = &logonHandle->rcvPacket->segm;
#endif

  if (ret == TCL_OK && messType == M_PARSE) {
    Tcl_DStringInit (&parsId);
#if ADABAS_VERSION >= 62
    ret = getParsId (interp, rcvSegment, &parsId);
#else
    ret = getParsId (interp, &logonHandle->sqlPacket, &parsId);
#endif
    if (!ret) {
      AppendResult (interp, "No returning parsid???");
      return TCL_ERROR;
    }
    ret = TCL_OK;
  }

  /*
   * Set the resultcount, if wanted (update, delete or insert).
   */

  if (messType == M_PARSE) {
    AdabasUnsetMsg  (interp, AdamsgRows);
  } else {
    if (withResCount) {
#if ADABAS_VERSION >= 62
      resCnt = getResultCount (rcvSegment);
#else /* ADABAS_VERSION <= 61 */
      resCnt = getResultCount (&logonHandle->sqlPacket);
#endif
    } else {
      resCnt = selectKind == intoSelect ? 1 : 0;
    }
    AdabasSetMsgInt (interp, AdamsgRows, resCnt);
  }
  AdabasUnsetMsg  (interp, AdamsgIntoVars);

  if (withInfo) {
#if ADABAS_VERSION >= 62
    if (!getResultTable (rcvSegment, cursorVars->resultTable)) {
      *cursorVars->resultTable = 0;
    }
#else /* ADABAS_VERSION <= 61 */
    if (!getResultTable (&logonHandle->sqlPacket, messType,
			 cursorVars->resultTable)) {
      if (selectKind == showSelect || selectKind == explainSelect) {
	regExp = Tcl_RegExpCompile (interp,
				    "^[ \t\n]*[a-zA-Z]*[ \t\n]*\\(([^)]*)\\)");
	if (Tcl_RegExpExec (interp, regExp, cmd, cmd)) {
	  Tcl_RegExpRange (regExp, 1, &start, &end);
	  if (end-start > 40) {
	    end = start+40;
	  }
	  p2c (cursorVars->resultTable, start, end-start);
	  for (start = cursorVars->resultTable; *start; start++) {
	    *start = toupper (*start);
	  }
	} else {
	  strcpy (cursorVars->resultTable, "SHOW");
	}
      } else if (selectKind == massSelect && resultTable && *resultTable) {
	strncpy (cursorVars->resultTable, resultTable,
		 sizeof (cursorVars->resultTable));
      } else {
	cursorVars->resultTable[0] = 0;
      }
    }
#endif
  } else {
    cursorVars->cntParams = 0;
  }

#if ADABAS_VERSION >= 62
  describeNeeded = withInfo &&  messType == M_PARSE;
#else
  describeNeeded = withInfo && (messType == M_PARSE || sqlModeInt >= 0);
#endif
  if (describeNeeded && ret == TCL_OK) {
    Tcl_DStringInit (&describe);
    Tcl_DStringAppend (&describe, "DESCRIBE ", -1);
    Tcl_DStringAppend (&describe, cursorVars->resultTable, -1);
    ret = AdabasSend2PartPacket (interp, logonHandle,
				 pkCommand, Tcl_DStringValue (&describe), 0,
				 pkNil, "", 0);
    Tcl_DStringFree (&describe);

#if ADABAS_VERSION >= 62
    rcvSegment = &logonHandle->rcvPacket->segm;
#endif
  }

  if (withInfo && ret == TCL_OK) {
#if ADABAS_VERSION >= 62
    cursorVars->cntParams = getShortInfos (rcvSegment,
					   &cursorVars->shortInfos,
					   &cursorVars->columnNames);
#else
    cursorVars->cntParams = getShortInfos (&logonHandle->sqlPacket,
					   &cursorVars->shortInfos,
					   &cursorVars->columnNames);
#endif
  }

  /*
   * First clean up the resources before leaving the procedure.
   */

  if (selectKind == massSelect && resultTable && *resultTable) {
    Tcl_DStringFree (&cmdStr);
  }
  if (sqlModeInt >= 0) {
    logonHandle->packetInfo.sqlMode = oldSqlMode;
  }

  /*
   * If we modified the sql command (put a DECLARE ... CURSOR FOR
   * in front of it), we have to adjust the error position.
   */

  if (ret == TCL_ERROR) {
    if (errorOffset
	&& AdabasGetMsgInt (interp, AdamsgErrorpos, &pos) == TCL_OK
	&& pos > errorOffset) {
      AdabasSetMsgInt (interp, AdamsgErrorpos, pos - errorOffset);
    }
    return TCL_ERROR;
  }

  if (messType == M_PARSE) {
    
    if (!withParameter) {
      AppendResult (interp, Tcl_DStringValue (&parsId));
      Tcl_DStringFree (&parsId);
      return TCL_OK;
    }

    ret = executeWithParameter (interp, cursorVars, &parsId);
    Tcl_DStringFree (&parsId);
    return ret;
  }

  switch (selectKind) {
  case intoSelect: selectInto (interp, cursorVars, selInto); break;
  case massSelect: invalidateResultTable (cursorVars);       break;
  default:                                                   break;
  }

#if ADABAS_VERSION <= 61
  if (sqlModeInt >= 0 && logonHandle->packetInfo.commitImmediately) {
    return AdabasSendCmdPacket (interp, logonHandle, 0, 1,
				"COMMIT WORK");
  }
#endif
  
  return TCL_OK;
}

/* -------------------------------------------------------------- */

static int
selectInto (interp, cursorVars, selectInto)
     Tcl_Interp *interp;		/* Current interpreter. */
     CursorInfo *cursorVars;		/* Current cursor. */
     char       *selectInto;		/* Pointing to first parameter. */
{
  Tcl_RegExp  regExp;
  char       *start;
  char       *end;
  char       *data;
  char       *nullValue;
  char       *specialNull;
  char       *colon  = selectInto;
  int         paramI = 0;
  int         ret    = TCL_OK;
  field_info *currSi;
#ifdef HAS_TCL_OBJECTS
  Tcl_Obj    *varObj;
  Tcl_Obj    *resObj;
  int         dummyLength;		/* Out parameter of GetXxxFromObj. */
#else
  Tcl_DString colName;
#endif

  if (!(data = getDataOfCursor (interp, cursorVars, 0)))
    return TCL_ERROR;

  regExp = Tcl_RegExpCompile (interp, ":([a-zA-Z0-9_()]*)");

  currSi = cursorVars->shortInfos;
  getNullValue (interp, &nullValue, &specialNull);

  while (colon && Tcl_RegExpExec (interp, regExp, colon, selectInto)) {
    Tcl_RegExpRange (regExp, 1, &start, &end);

#ifdef HAS_TCL_OBJECTS

    varObj = Tcl_NewStringObj (start, end-start);
    resObj = Tcl_NewObj ();
    ret    = appendOneValue (interp, data, resObj, nullValue, specialNull,
			     currSi, &cursorVars->longColFound, 0);
    Tcl_ObjSetVar2 (interp, varObj, (Tcl_Obj *) NULL, resObj,
		    TCL_PART1_NOT_PARSED);
    
    AdabasLAppendMsg (interp, AdamsgIntoVars,
		      Tcl_GetStringFromObj (interp, varObj, &dummyLength));
    Tcl_DecrRefCount (varObj);
    Tcl_DecrRefCount (resObj);

#else

    Tcl_DStringInit   (&colName);
    Tcl_DStringAppend (&colName, start, end-start);
    ret = appendOneValue (interp, data, (Tcl_Obj *) NULL, nullValue,
			  specialNull, currSi, &cursorVars->longColFound, 0);
    Tcl_SetVar (interp, Tcl_DStringValue (&colName), interp->result, 0);
    Tcl_ResetObjResult (interp);
    AdabasLAppendMsg (interp, AdamsgIntoVars, Tcl_DStringValue (&colName));
    Tcl_DStringFree  (&colName);

#endif

    colon = (char *) NULL;
    if (++paramI < cursorVars->cntParams) {
      while (*end == ' ') end++;
      if (*end == ',') {
	end++; 
	while (*end == ' ') end++;
	if (*end == ':') {
	  colon = end;
	  currSi++;
	}
      }
    }
  }

  return ret;
}

/* -------------------------------------------------------------- */

static void
invalidateResultTable (cursorVars)
     CursorInfo *cursorVars;		/* Current cursor. */
{
  Tcl_HashEntry *hashEntry;
  Tcl_HashSearch hashSearch;
  CursorInfo    *cmpCursor;

  hashEntry = Tcl_FirstHashEntry (&cursorHash, &hashSearch);
  while (hashEntry) {
    cmpCursor = (CursorInfo *) Tcl_GetHashValue (hashEntry);

    /*
     * If we have another cursor for the same logon with the same
     * resulttablename, its cursor infos will be invalidated for future
     * fetches, since this information is now obsolete.
     */

    if (cursorVars != cmpCursor
	&& cursorVars->logonInfo == cmpCursor->logonInfo
	&& !strcmp (cursorVars->resultTable, cmpCursor->resultTable))
      cmpCursor->cntParams = 0;

    hashEntry = Tcl_NextHashEntry (&hashSearch);
  }
}

/* -------------------------------------------------------------- */

int
AdabasExecParsId (interp, cursorVars, parsId)
     Tcl_Interp *interp;		/* Current interpreter. */
     CursorInfo *cursorVars;		/* Current cursor. */
     char       *parsId;		/* ParsId to execute. */
{
  int orgMessType;
  int ret;

  Tcl_ResetObjResult (interp);
  if (checkCursor (interp, cursorVars) == TCL_ERROR) {
    return TCL_ERROR;
  }

  orgMessType = cursorVars->logonInfo->packetInfo.messType;
  cursorVars->logonInfo->packetInfo.messType = M_EXECUTE;

  ret = AdabasSend2PartPacket (interp, cursorVars->logonInfo,
			       pkParsid, parsId, strlen (parsId),
			       pkNil, "", 0);

  cursorVars->logonInfo->packetInfo.messType = orgMessType;

  return ret;
}

/* -------------------------------------------------------------- */

static int
executeWithParameter (interp, cursorVars, parsId)
     Tcl_Interp  *interp;		/* Current interpreter. */
     CursorInfo  *cursorVars;		/* Current cursor. */
     Tcl_DString *parsId;		/* ParsId to execute. */
{
  int orgMessType;
  int ret;

  orgMessType = cursorVars->logonInfo->packetInfo.messType;
  cursorVars->logonInfo->packetInfo.messType = M_EXECUTE;

  ret = AdabasSend2PartPacket (interp, cursorVars->logonInfo,
			       pkParsid, Tcl_DStringValue (parsId),
			       Tcl_DStringLength (parsId),
			       pkNil, "", 0);

  cursorVars->logonInfo->packetInfo.messType = orgMessType;
  if (ret == TCL_ERROR) return TCL_ERROR;

  return AdabasSend2PartPacket (interp, cursorVars->logonInfo,
				pkCommand, "DROP PARSEID", 0,
				pkParsid, Tcl_DStringValue (parsId),
				Tcl_DStringLength (parsId));
}

/* -------------------------------------------------------------- */

int
AdabasDropParsId (interp, cursorVars, parsId)
     Tcl_Interp *interp;		/* Current interpreter. */
     CursorInfo *cursorVars;		/* Current cursor. */
     char       *parsId;		/* ParsId to drop. */
{
  Tcl_ResetObjResult (interp);
  if (checkCursor (interp, cursorVars) == TCL_ERROR) {
    return TCL_ERROR;
  }

  return AdabasSend2PartPacket (interp, cursorVars->logonInfo,
				pkCommand, "DROP PARSEID", 0,
				pkParsid, parsId, strlen (parsId));
}

/* -------------------------------------------------------------- */

int
AdabasCancel (interp, cursorVars)
     Tcl_Interp *interp;		/* Current interpreter. */
     CursorInfo *cursorVars;		/* Current cursor. */
{
  Tcl_ResetObjResult (interp);
  if (checkCursor (interp, cursorVars) == TCL_ERROR) {
    return TCL_ERROR;
  }

  cursorVars->cntParams    = 0;
  cursorVars->longColFound = 0;
  AdabasRteCancel (&cursorVars->logonInfo->rteInfo);
  return TCL_OK;
}

/* -------------------------------------------------------------- */

int
AdabasFetch (interp, cursorVars, target, count, sqlMode, arraySet)
     Tcl_Interp *interp;		/* Current interpreter. */
     CursorInfo *cursorVars;		/* Current cursor. */
     char       *target;		/* Target to fetch (e.g. "first"). */
     int         count;			/* Count for mass fetches. */
     char       *sqlMode;		/* Sqlmode (or NULL). */
     int         arraySet;		/* Return results sa column/value pairs? */
{
  char        cmdBuf[100];
  char        buf[20];
  int         pos;
  int         ix;
  long        rc;
  char       *data;
  char       *nullValue;
  char       *specialNull;
  int         ret;
  int         oldSqlMode  = -1;
  field_info *currSi;
  char       *currCol    = NULL;
  int         sqlModeInt = -1;
  int         length;			/* Length (used for option parsing). */
  int         c;			/* Char   (used for option parsing). */
  Tcl_Obj    *resObj = NULL;
#ifdef HAS_TCL_OBJECTS
  Tcl_Obj    *resList;
#else
  char        currName[65];
#endif

  Tcl_ResetObjResult (interp);
  if (checkCursor (interp, cursorVars) == TCL_ERROR) {
    return TCL_ERROR;
  }

  if (!target) target = "";
  if (*target) {
    c      = target[0];
    length = strlen (target);
    if (c == 'f' && !strncmp (target, "first", length))
      target = "FIRST";
    else if (c == 'l' && !strncmp (target, "last",  length))
      target = "LAST";
    else if (c == 'n' && !strncmp (target, "next",  length))
      target = "NEXT";
    else if (c == 'p' && !strncmp (target, "prev",  length))
      target = "PREV";
    else if (Tcl_GetInt ((Tcl_Interp *) NULL, target, &pos) == TCL_OK) {
      sprintf (buf, "POS (%d)", pos);
      target = buf;
    } else {
      AppendResult (interp, "bad value for switch -position \"");
      AppendResult (interp, target);
      AppendResult (interp, "\": must be -first, -last, -prev, -next or a ");
      AppendResult (interp, "number");
      return TCL_ERROR;
    }
  }

  if (sqlMode && *sqlMode) {
    sqlModeInt = scanSqlMode (interp, sqlMode, 0);
#if ADABAS_VERSION >= 62
    if (sqlModeInt == SM_NIL) return TCL_ERROR;
#else /* ADABAS_VERSION <= 61 */
    if (sqlModeInt < 0) return TCL_ERROR;
#endif
  }

  if (sqlModeInt >= 0) {
    oldSqlMode = cursorVars->logonInfo->packetInfo.sqlMode;
    cursorVars->logonInfo->packetInfo.sqlMode = sqlModeInt;
  }

  /*
   * Send the fetch command to the database kernel.
   */

  sprintf (cmdBuf, "FETCH %s %s INTO ?", target, cursorVars->resultTable);
  ret = AdabasSendCmdPacket (interp, cursorVars->logonInfo,
			     0, count, cmdBuf);
  if (sqlModeInt >= 0)
    cursorVars->logonInfo->packetInfo.sqlMode = oldSqlMode;

  if (ret == TCL_ERROR) {

    /*
     * Check, if a positional fetch was called; then a better error
     * message then MISSING KEYWORD or UNKNOWN RESULT TABLE:LAST
     * should be given.
     */

    if (*target
	&& AdabasGetMsgInt (interp, AdamsgRc, &rc) == TCL_OK
	&& (rc == -4000 || rc == -3008 || rc == -5015)) {
      SetResult (interp, "Positional fetch only valid in sqlmode adabas");
      return TCL_ERROR;
    }
    if (AdabasGetMsgInt (interp, AdamsgRc, &rc) == TCL_OK
	&& rc != 100) {
      return TCL_ERROR;
    }
    Tcl_ResetObjResult (interp);
    return TCL_OK;
  }

  if (!(data = getDataOfCursor (interp, cursorVars,
				count > 1 ? MASS_CMD : 0)))
    return TCL_ERROR;


#ifdef HAS_TCL_OBJECTS
  resList = Tcl_NewListObj (0, (Tcl_Obj **) NULL);
  Tcl_SetObjResult (interp, resList);
#endif

  getNullValue (interp, &nullValue, &specialNull);
  while (data && ret == TCL_OK) {

    if (count > 1) {
#ifdef HAS_TCL_OBJECTS
      resList = Tcl_NewListObj (0, (Tcl_Obj **) NULL);
      Tcl_ListObjAppendElement (interp, Tcl_GetObjResult (interp), resList);
#else
      AppendResult (interp, "{");
#endif
    }

    currSi  = cursorVars->shortInfos;
    if (arraySet)
      currCol = cursorVars->columnNames;
    for (ix = 0; ix < cursorVars->cntParams && ret == TCL_OK; ix++) {
      if (arraySet) {
#ifdef HAS_TCL_OBJECTS
	Tcl_ListObjAppendElement (interp, resList,
				  Tcl_NewStringObj (currCol+1, *currCol));
#else
	sprintf (currName, "%.*s", *currCol, currCol+1);
	Tcl_AppendElement (interp, currName);
#endif
	currCol += *currCol + 1;
      }
#ifdef HAS_TCL_OBJECTS
      resObj = Tcl_NewObj ();
      Tcl_ListObjAppendElement (interp, resList, resObj);
#endif
      ret = appendOneValue (interp, data, resObj, nullValue, specialNull,
			    currSi++, &cursorVars->longColFound, 1);
    }
    if (count > 1) {
#ifndef HAS_TCL_OBJECTS
      AppendResult (interp, "} ");
#endif
      data = getNextData (interp, data, cursorVars);
    } else {
      data = NULL;
    }
  }

  return ret;
}

/* -------------------------------------------------------------- */

static int
appendOneValue (interp, data, destObj, nullValue, specialNull,
		currSi, longColFound, isElement)
     Tcl_Interp *interp;		/* Current interpreter. */
     char       *data;			/* Value to append as string. */
     Tcl_Obj    *destObj;		/* Destination object to append element,
					 * or Null, if dest is interp->result. */
     char       *nullValue;		/* Representation of a NULL value. */
     char       *specialNull;		/* Representation of a SPECIAL NULL. */
     field_info *currSi;		/* Shortfield info of value. */
     int        *longColFound;		/* Set to 1, if value is LONG. */
     int         isElement;		/* append result as element? */
{
  int ret;

#if ADABAS_VERSION >= 62
  if (currSi->io_type == IO_INPUT)
    return TCL_OK; /* nothing to append. */
#endif    

  ret = AdabasAppendData (interp, data+currSi->bufpos-1, destObj,
			  currSi->data_type, currSi->in_out_len,
			  currSi->length, currSi->frac,
			  nullValue, specialNull, isElement);

#if ADABAS_VERSION >= 62
  if (currSi->data_type == DSTRA  ||
      currSi->data_type == DSTRE  ||
      currSi->data_type == DSTRB  ||
      currSi->data_type == DSTRDB ||
      currSi->data_type == DSTRUNI)
    *longColFound = 1;
#else /* ADABAS_VERSION <= 61 */
  if (currSi->data_type == CSP_INFO_C_OLDLONG_CHAR  ||
      currSi->data_type == CSP_INFO_B_OLDLONG_BYTE  ||
      currSi->data_type == CSP_INFO_A_OLDLONG_ASCII_DBYTE)
    *longColFound = 1;
#endif    

  return ret;
}

/* -------------------------------------------------------------- */

int
AdabasFetchAll (interp, cursorName, count, tclCommand)
     Tcl_Interp *interp;		/* Current interpreter. */
     char       *cursorName;		/* Name of current cursor. */
     int         count;			/* Count for mass fetches. */
     char       *tclCommand;		/* Command to execute for each row. */
{
  char *argv[5];
  char *command;
  int ret;
  char countBuf[30];

  Tcl_ResetObjResult (interp);
  if (!AdabasCursorHandle (interp, cursorName)) {
    return TCL_ERROR;
  }

  sprintf (countBuf, "%d", count);

  argv[0] = "_adafetch_all";
  argv[1] = cursorName;
  argv[2] = countBuf;
  argv[3] = tclCommand;
  argv[4] = (char *) NULL;

  command = Tcl_Merge (4, argv);
  ret     = Tcl_Eval (interp, command);
  ckfree (command);
  return ret;
}

/* -------------------------------------------------------------- */

static char *
getNextData (interp, data, cursorVars)
     Tcl_Interp *interp;		/* Current interpreter. */
     char       *data;			/* Buffer of data part. */
     CursorInfo *cursorVars;		/* Current cursor. */
{
  field_info *lastSi;
  int4        nextOff;

  cursorVars->rowIx++;
  lastSi  = cursorVars->shortInfos+(cursorVars->cntParams-1);
  nextOff = lastSi->bufpos+lastSi->in_out_len-1;

  if (cursorVars->rowIx >= cursorVars->massRows ||
      cursorVars->dataEnd + nextOff >= cursorVars->dataLength) {

    /*
     * We reached the end of data, so better return a NULL.
     */

    return NULL;
  }

  /*
   * Increase the dataEnd marker in cursorVars and
   * return the adjusted data pointer.
   */

  cursorVars->dataEnd += nextOff;
  return data+nextOff;
}

/* -------------------------------------------------------------- */

int
AdabasColumns (interp, cursorVars)
     Tcl_Interp *interp;		/* Current interpreter. */
     CursorInfo *cursorVars;		/* Current cursor. */
{
  char        number[20];
  int         ix;
  char       *dataType;
  int         length;
  int         frac;
  field_info *currSi  = cursorVars->shortInfos;
  char       *currCol = cursorVars->columnNames;
#ifndef HAS_TCL_OBJECTS
  char        name[65];
#endif

  Tcl_ResetObjResult (interp);
  if (checkCursor (interp, cursorVars) == TCL_ERROR) {
    return TCL_ERROR;
  }

  AdabasUnsetMsg (interp, AdamsgColtypes);
  AdabasUnsetMsg (interp, AdamsgCollengths);
  AdabasUnsetMsg (interp, AdamsgColprecs);
  AdabasUnsetMsg (interp, AdamsgColscales);

  for (ix = 0; ix < cursorVars->cntParams; ix++) {
    dataType = dataType2String (currSi->data_type);
    length   = currSi->length;
    frac     = currSi->frac;

    AdabasLAppendMsg (interp, AdamsgColtypes,   dataType);
    sprintf (number, "%d", length);
    AdabasLAppendMsg (interp, AdamsgCollengths, number);
    AdabasLAppendMsg (interp, AdamsgColprecs,   number);
    sprintf (number, "%d", frac);
    AdabasLAppendMsg (interp, AdamsgColscales,  number);

#ifdef HAS_TCL_OBJECTS
    Tcl_ListObjAppendElement (interp, Tcl_GetObjResult (interp),
			      Tcl_NewStringObj (currCol+1, *currCol));
#else
    sprintf (name, "%.*s", *currCol, currCol+1);
    Tcl_AppendElement (interp, name);
#endif

    currSi++;
    currCol += *currCol + 1;
  }
  return TCL_OK;
}

/* -------------------------------------------------------------- */

int
AdabasReadLong (interp, cursorVars, descrString, tableName, columnName,
		whereCond, encoding, fileName)
     Tcl_Interp  *interp;		/* Current interpreter. */
     CursorInfo  *cursorVars;		/* Current cursor. */
     char        *descrString;		/* Long descriptor (or NULL). */
     char        *tableName;		/* Table name (or NULL). */
     char        *columnName;		/* Column name (or NULL). */
     char        *whereCond;		/* Where coindition (or NULL). */
     int          encoding;		/* Encoding kind. */
     char        *fileName;		/* Name of file to write in (or NULL). */
{
  long_descriptor longDesc;
  long            bytes;
  int             cont;
  char           *nullValue;
  char           *specialNull;

  Tcl_ResetObjResult (interp);
  if (checkCursor (interp, cursorVars) == TCL_ERROR) {
    return TCL_ERROR;
  }

  if (descrString) {

    /*
     * The variant with the long descriptor is given.
     * First check, if it is a nullvalue.
     */

    getNullValue (interp, &nullValue, &specialNull);
    if (!strcmp (descrString, nullValue)) {
      AppendResult (interp, nullValue);
      return TCL_OK;
    }
    if (list2LongDescr (interp, descrString, &longDesc) == TCL_ERROR) {
      return TCL_ERROR;
    }
    if (!cursorVars->longColFound) {
      SetResult (interp, "call of adareadlong must be preceded by adafetch");
      return TCL_ERROR;
    }

  } else {

    /*
     * The variant with table, column and condition is given.
     * So first get a new copy of the cursor handle, then start the SELECT.
     */

    if (!(cursorVars = AdabasOpen (interp, cursorVars->logonInfo))) {
      return TCL_ERROR;
    }
    if (AdabasLongSelect (interp, cursorVars, tableName, columnName,
			  whereCond, &longDesc) == TCL_ERROR) {
      (void) AdabasClose (interp, cursorVars);
      return TCL_ERROR;
    }
  }

  if (longDesc.valmode == VM_NODATA) {
    bytes = 0;
    cont  = 1;
  } else {
    if (AdabasLongReadFirst (interp, cursorVars, &longDesc, encoding,
			     fileName, &bytes, &cont) == TCL_ERROR) {
      if (!descrString) {
	(void) AdabasClose (interp, cursorVars);
      }
      return TCL_ERROR;
    }
  }

  while (cont) {
    if (AdabasLongReadNext (interp, cursorVars, &longDesc, encoding,
			    fileName, &bytes, &cont) == TCL_ERROR) {
      if (!descrString) {
	(void) AdabasClose (interp, cursorVars);
      }
      return TCL_ERROR;
    }
  }

  /*
   * If called with a filename to write in, the number of
   * bytes written are returned.
   */

  if (fileName) {
    SetIntResult (interp, bytes);
  }

  if (!descrString) {
    (void) AdabasClose (interp, cursorVars);
  }
  return TCL_OK;
}

/* -------------------------------------------------------------- */

int
AdabasWriteLong (interp, cursorVars, tableName, columnName,
		 whereCond, encoding, fileName, longValue)
     Tcl_Interp  *interp;		/* Current interpreter. */
     CursorInfo  *cursorVars;		/* Current cursor. */
     char        *tableName;		/* Table name (or NULL). */
     char        *columnName;		/* Column name (or NULL). */
     char        *whereCond;		/* Where coindition (or NULL). */
     int          encoding;		/* Encoding kind. */
     char        *fileName;		/* Name of file to read from (or NULL). */
     char        *longValue;		/* Value to write (or NULL). */
{
  long_descriptor longDesc;
  int             longLength;
  int4            bytes;
  int             cont;
  Tcl_DString     parsId;
  Tcl_DString     substVal;

  Tcl_ResetObjResult (interp);
  if (checkCursor (interp, cursorVars) == TCL_ERROR) {
    return TCL_ERROR;
  }

  if (!(cursorVars = AdabasOpen (interp, cursorVars->logonInfo))) {
    return TCL_ERROR;
  }

  /*
   * Send the constructed UPDATE sql command to the kernel for parsing.
   */

  if (AdabasLongUpdate (interp, cursorVars, tableName, columnName,
			whereCond, &parsId) == TCL_ERROR) {
    (void) AdabasClose (interp, cursorVars);
    return TCL_ERROR;
  }

  if (AdabasLongValLength (interp, fileName, longValue, encoding,
			   &substVal, &longLength) == TCL_ERROR) {
    (void) AdabasClose (interp, cursorVars);
    return TCL_ERROR;
  }
    
  /*
   * Now the parsid will be executed, the first part of the long value
   * (or all, if it fits into the packet) is given as data.
   */

  if (AdabasLongWriteFirst (interp, cursorVars, &longDesc, &parsId, fileName,
			    Tcl_DStringValue (&substVal), encoding, longLength,
			    &bytes, &cont) == TCL_ERROR) {
    Tcl_DStringFree (&parsId);
    if (longValue) {
      Tcl_DStringFree (&substVal);
    }
    (void) AdabasClose (interp, cursorVars);
    return TCL_ERROR;
  }

  if (cont) {

    /*
     * While not all of the long value is written,
     * continue with sending PUTVAL packets.
     */

    while (cont) {
      if (AdabasLongWriteNext (interp, cursorVars, &longDesc, fileName,
			       Tcl_DStringValue (&substVal), encoding,
			       longLength, &bytes, &cont) == TCL_ERROR) {
	Tcl_DStringFree (&parsId);
	if (longValue) Tcl_DStringFree (&substVal);
	(void) AdabasClose (interp, cursorVars);
	return TCL_ERROR;
      }
    }
  }

  /*
   * Finally throw away the generated parsid; ignore its return value,
   * since we are in the clean up phase now.
   */

  (void) AdabasSend2PartPacket (interp, cursorVars->logonInfo,
				pkCommand, "DROP PARSEID", 0,
				pkParsid, Tcl_DStringValue (&parsId),
				Tcl_DStringLength (&parsId));

  Tcl_DStringFree (&parsId);
  if (longValue) Tcl_DStringFree (&substVal);
  return AdabasClose (interp, cursorVars);
}

/* -------------------------------------------------------------- */

int
AdabasAutoCommit (interp, vars, onOff)
     Tcl_Interp *interp;		/* Current interpreter. */
     AdabasInfo *vars;			/* Logon handle. */
     int         onOff;			/* autocommit on? */
{
  Tcl_ResetObjResult (interp);
  if (AdabasCheckLogon (interp, vars) == TCL_ERROR) {
    return TCL_ERROR;
  }

  vars->packetInfo.commitImmediately = onOff;
  return TCL_OK;
}

/* -------------------------------------------------------------- */

int
AdabasCommit (interp, vars)
     Tcl_Interp *interp;		/* Current interpreter. */
     AdabasInfo *vars;			/* Logon handle. */
{
  Tcl_ResetObjResult (interp);
  if (AdabasCheckLogon (interp, vars) == TCL_ERROR) {
    return TCL_ERROR;
  }

  return AdabasSendCmdPacket (interp, vars, 0, 1, "COMMIT WORK");
}

/* -------------------------------------------------------------- */

int
AdabasRollback (interp, vars)
     Tcl_Interp *interp;		/* Current interpreter. */
     AdabasInfo *vars;			/* Logon handle. */
{
  Tcl_ResetObjResult (interp);
  if (AdabasCheckLogon (interp, vars) == TCL_ERROR) {
    return TCL_ERROR;
  }

  return AdabasSendCmdPacket (interp, vars, 0, 1, "ROLLBACK WORK");
}

/* -------------------------------------------------------------- */

int
AdabasUtil (interp, vars, cmd)
     Tcl_Interp *interp;		/* Current interpreter. */
     AdabasInfo *vars;			/* Logon handle. */
     char       *cmd;			/* Utility command to execute. */
{
  char       *data;
  int4        dataLength;
  int         argCount;
  int         cntParams;
  int         ret;
  int         ix;
  int         longColFound;
  char       *nullValue;
  char       *specialNull;
  char        dbName[200]; /* more than enough. */  
  field_info *currSi;
  field_info *shortInfos  = NULL;
  char       *columnNames = NULL;
  Tcl_Obj    *resObj;
#ifndef HAS_TCL_OBJECTS
  Tcl_DString resultStr;
#endif

  Tcl_ResetObjResult (interp);
  if (AdabasCheckLogon (interp, vars) == TCL_ERROR)
    return TCL_ERROR;

  if (vars->packetInfo.messType != M_UTILITY) {
    SetResult (interp, "Logon handle must be utility service");
    return TCL_ERROR;
  }

  if (AdabasSendCmdPacket (interp, vars, 0, 1, cmd) == TCL_ERROR)
    return TCL_ERROR;

#if ADABAS_VERSION >= 62
  data = getData (&vars->rcvPacket->segm, &dataLength, &argCount, UTILITY_CMD);
#else /* ADABAS_VERSION <= 61 */
  data = getData (&vars->sqlPacket, &dataLength, &argCount, UTILITY_CMD);
#endif
  if (data && dataLength) {
#if ADABAS_VERSION >= 62
    cntParams = getShortInfos (&vars->rcvPacket->segm, &shortInfos, &columnNames);
#else /* ADABAS_VERSION <= 61 */
    cntParams = 0;
#endif

    if (cntParams && shortInfos) {
      getNullValue (interp, &nullValue, &specialNull);
      currSi = shortInfos;
      ret    = TCL_OK;
#ifdef HAS_TCL_OBJECTS
      resObj = Tcl_GetObjResult (interp);
#else
      resObj = (Tcl_Obj *) NULL;
#endif
      for (ix = 0; ix < cntParams && ret == TCL_OK; ix++) {
	ret = appendOneValue (interp, data, resObj, nullValue,
			      specialNull, currSi++, &longColFound, 1);
      }

    } else 
      switch (*data) {
      case  2:
	p2c (dbName, data+1, dataLength-1);
	AppendResult (interp, "warm ");
	AppendResult (interp, dbName);
	break;
      case 0: case 1:
	AppendResult (interp, "cold");
	break;
      default:
#ifdef HAS_TCL_OBJECTS
	Tcl_SetStringObj (Tcl_GetObjResult (interp), data, dataLength);
#else
	Tcl_DStringInit   (&resultStr);
	Tcl_DStringAppend (&resultStr, data, dataLength);
	Tcl_DStringResult (interp, &resultStr);
#endif
    }
    if (shortInfos)  ckfree ((char *) shortInfos);
    if (columnNames) ckfree ((char *) columnNames);
  }
  return TCL_OK;
}

/* -------------------------------------------------------------- */

int
AdabasSpecial (interp, vars, messType, data)
     Tcl_Interp *interp;		/* Current interpreter. */
     AdabasInfo *vars;			/* Logon handle. */
     int         messType;		/* Special message type. */
     char       *data;			/* Data for special command (or ""). */
{
  int ret;
  int orgMessType = vars->packetInfo.messType;

  Tcl_ResetObjResult (interp);
  if (AdabasCheckLogon (interp, vars) == TCL_ERROR) {
    return TCL_ERROR;
  }

  vars->packetInfo.messType = messType;
  ret = AdabasSend2PartPacket (interp, vars, pkData, data, 0, pkNil, "", 0);
  vars->packetInfo.messType = orgMessType;

  return ret;
}

/* -------------------------------------------------------------- */

AdabasInfo *
AdabasLogon (interp, serverdb, connectStr, isolation, sqlMode, serv, applKind)
     Tcl_Interp *interp;		/* Current interpreter. */
     char       *serverdb;		/* Serverdb to connect to (or NULL) */
     char       *connectStr;		/* Connect string (e.g. user,passwd) */
     int         isolation;		/* Isolation level (or -1) */
     char       *sqlMode;		/* Wanted Sqlmode */
     int         serv;			/* SQL Service (user/utility) */
     int         applKind;  	 	/* User, control or odbs service */
{
  Tcl_RegExp   regExp;
  Tcl_DString  connect;
  Tcl_DString  data;
  Tcl_DString  cUser;
  Tcl_DString  cCryptPasswd;
  char         isoString[80];
  Tcl_DString  help;
  char         xCryptPasswd[24];
  char         xUser[sizeof(username)+1];
  char         xNode[sizeof(nodeid)+1];
  char         xDbName[sizeof(dbname)+1];
  AdabasInfo  *vars;
  char        *xuserKey;
  char        *start;
  char        *end;
  int          ret;
  char        *user;
  xuser_record userParams;

  Tcl_ResetObjResult (interp);

  if (isolation >= 0
      && (isolation != 0 && isolation != 15 &&
	  isolation != 1 && isolation != 10 &&
	  isolation != 2 && isolation != 20 &&
	  isolation != 3 && isolation != 30)) {
    SetResult (interp, "isolation must be 0, 1, 2, 3, 10, 15, 20 or 30");
    return (AdabasInfo *) NULL;
  }
  if (sqlMode && scanSqlMode (interp, sqlMode, 1) < 0)
    return (AdabasInfo *) NULL;
    
  Tcl_DStringInit (&help);

  regExp = Tcl_RegExpCompile (interp, "^([^,]*),(.*)$");
  if (!Tcl_RegExpExec (interp, regExp, connectStr, connectStr)) {

    /*
     * There is a pseudo user "-noconnect" to allow sessions without
     * a real connect (e.g. for vtrace or switch commands).
     */

    if (!strcmp (connectStr, "-noconnect")) {
      if (sqlMode) {
	SetResult (interp, "specification of sqlmode useless, if not connecting");
	return (AdabasInfo *) NULL;
      }
      user = (char *) NULL; /* special flag: no sql connection, please. */
    } else {
      AppendResult (interp, "Invalid connectString \"");
      AppendResult (interp, connectStr);
      AppendResult (interp, "\", must be \"user,passwd\", \",userKey\", \",\"");
      AppendResult (interp, " or \"-noconnect\"");
      return (AdabasInfo *) NULL;
    }
  } else {
    Tcl_DStringInit (&cUser);
    Tcl_DStringInit (&cCryptPasswd);

    Tcl_RegExpRange (regExp, USER_RANGE, &start, &end);
    if (start == end) {

      /*
       * There was no user given; perhaps the name behind
       * the comma is meant as an xuser key.
       */

      Tcl_RegExpRange (regExp, PASSWD_RANGE, &start, &end);
      if (start == end) {

	/*
	 * There was no xuser key given either.
	 * So let's look after the default user.
	 */

	xuserKey = NULL;
      } else {
	Tcl_DStringAppend (&help, start, end-start);
	xuserKey = Tcl_DStringValue (&help);
      }
      if (AdabasXUserGet (interp, xuserKey, &userParams) == TCL_ERROR) {
	Tcl_DStringFree (&help);
	return (AdabasInfo *) NULL;
      }

      p2c (xUser, userParams.xu_user,
	   sizeof (userParams.xu_user));
      p2c (xDbName, userParams.xu_serverdb,
	   sizeof (userParams.xu_serverdb));
      p2c (xNode, userParams.xu_servernode,
	   sizeof (userParams.xu_servernode));
      memcpy (xCryptPasswd, userParams.xu_password,
	      sizeof (userParams.xu_password));

      Tcl_DStringFree (&help);
      decodeSequence (interp, encEscape, (Tcl_Obj *) NULL, &cCryptPasswd,
		      xCryptPasswd, sizeof (xCryptPasswd));
      user = xUser;
      if (*xDbName && *xDbName != ' ') {
	if (*xNode && *xNode != ' ') {
	  Tcl_DStringAppend (&help, xNode, -1);
	  Tcl_DStringAppend (&help, ":",    1);
	}
	Tcl_DStringAppend (&help,  xDbName, -1);
	serverdb = Tcl_DStringValue (&help);
      }
    } else {

      /*
       * There was really a pair of user and password given.
       */

      Tcl_DStringAppend (&cUser, start, end-start); 
      user = Tcl_DStringValue (&cUser);

      Tcl_RegExpRange   (regExp, PASSWD_RANGE, &start, &end);
      Tcl_DStringAppend (&help, start, end-start);

      /*
       * Change passwd into uppercase before crypting it.
       */

      for (start = Tcl_DStringValue (&help); *start; start++)
	*start = toupper (*start);

      AdabasPasswdCrypt (interp, &cCryptPasswd, Tcl_DStringValue (&help));
    }
  }

  /*
   * Now open the connection to adabas with the given service and serverdb.
   */

  ret = AdabasOpenConnection (interp, serv, serverdb, "", &vars);
  Tcl_DStringFree (&help);
  if (ret == TCL_ERROR) return (AdabasInfo *) NULL;

  switch (applKind) {
  case ControlAppl:
    memcpy (vars->packetInfo.application, "CON", 3);
    break;
  case OdbcAppl:
    memcpy (vars->packetInfo.application, "ODB", 3);
    break;
  default:
    if (serv == SQL_UTILITY) {
      vars->packetInfo.messType = M_UTILITY;
    }
  }

  if (user) {

    /*
     * Now create a session with the given user, password (and sqlmode).
     */

    Tcl_DStringInit   (&connect);
    Tcl_DStringAppend (&connect, "CONNECT ", -1);
    Tcl_DStringAppend (&connect, user,       -1);
    Tcl_DStringAppend (&connect, " IDENTIFIED BY :x", -1);

    if (sqlMode) {
      Tcl_DStringAppend (&connect, " SQLMODE ", -1);
      Tcl_DStringAppend (&connect,   sqlMode,   -1);
    }

    if (isolation >= 0) {
      sprintf (isoString, " ISOLATION LEVEL %d", isolation);
      Tcl_DStringAppend (&connect, isoString, -1);
    }

    /*
     * Construct a data part, which consists of a defined byte,
     * the password and the terminal ID.
     */

    Tcl_DStringInit   (&data);
    decodeSequence    (interp, encEscape, (Tcl_Obj *) NULL, &data, "", 1);
    Tcl_DStringAppend (&data, Tcl_DStringValue (&cCryptPasswd), -1);
    Tcl_DStringAppend (&data, " ", 1);
    AdabasRteTermId   (Tcl_DStringValue (&help)); 
    Tcl_DStringAppend (&data, Tcl_DStringValue (&help), -1); 

    ret = AdabasSend2PartPacket (interp, vars,
				 pkCommand, Tcl_DStringValue (&connect), 0,
				 pkData,    Tcl_DStringValue (&data),    0);
  
    Tcl_DStringFree (&connect);
    Tcl_DStringFree (&data);
    Tcl_DStringFree (&help);
  }

  if (ret == TCL_ERROR) {
    (void) AdabasCloseConnection (interp, vars->logonName, vars);
    return (AdabasInfo *) NULL;
  }

  Tcl_ResetObjResult (interp);
  AppendResult (interp, vars->logonName);

  return vars;
}

/* -------------------------------------------------------------- */

int
AdabasLogoff (interp, vars)
     Tcl_Interp *interp;		/* Current interpreter. */
     AdabasInfo *vars;			/* Current logon. */
{
  Tcl_ResetObjResult (interp);
  if (AdabasCheckLogon (interp, vars) == TCL_ERROR) {
    return TCL_ERROR;
  }

  (void) AdabasSendCmdPacket (interp, vars, 0, 1, "ROLLBACK WORK RELEASE");
  (void) AdabasCloseConnection (interp, vars->logonName, vars);
  return TCL_OK;
}

/* -------------------------------------------------------------- */

static void
examineSqlCommand (interp, cmdString, withInfo, withResCount,
		   selectKind, selectInto)
     Tcl_Interp *interp;		/* Current interpreter. */
     char       *cmdString;		/* String containing a SQL command. */
     int        *withInfo;		/* SQL, that returns results? */
     int        *withResCount;		/* SQL, that returns rowcnt? */
     SelectKind *selectKind;		/* returned kind of sql statement. */
     char      **selectInto;		/* position of first parameter name
					 * in cmdString (for SELECT INTO). */
{
  Tcl_DString cmd;
  Tcl_RegExp  regExp;
  char       *start;
  char       *end;

  *withInfo     = 0;
  *withResCount = 0;
  *selectKind   = noSelect;
  *selectInto   = (char *) NULL;

  regExp = Tcl_RegExpCompile (interp, "^[ \t\n]*([^ \t\n(]*)");
  if (!Tcl_RegExpExec (interp, regExp, cmdString, cmdString)) return;

  Tcl_RegExpRange   (regExp, CMD_RANGE, &start, &end);
  Tcl_DStringInit   (&cmd);
  Tcl_DStringAppend (&cmd, start, end-start);

  for (start = Tcl_DStringValue (&cmd); *start; start++)
    *start = toupper (*start);

  if (!strcmp (Tcl_DStringValue (&cmd), "SELECT")) {
    *selectKind = massSelect;
    *withInfo   = 1;
    /*
     * Now that we found a SELECT, lets look if it is a SELECT INTO.
     */
    Tcl_DStringFree   (&cmd);
    Tcl_DStringAppend (&cmd, cmdString, -1);
    for (start = Tcl_DStringValue (&cmd); *start; start++)
      *start = toupper (*start);
    regExp = Tcl_RegExpCompile (interp, " INTO  *(:)");
    if (Tcl_RegExpExec (interp, regExp,
			Tcl_DStringValue (&cmd), Tcl_DStringValue (&cmd))) {
      Tcl_RegExpRange (regExp, 1, &start, &end);
      *selectKind = intoSelect;
      *selectInto = cmdString + (start - Tcl_DStringValue (&cmd));
    }

  } else if (!strcmp (Tcl_DStringValue (&cmd), "EXPLAIN")) {
    *withInfo     = 1;
    *selectKind   = explainSelect;

  } else if (!strcmp (Tcl_DStringValue (&cmd), "SHOW")) {
    *withInfo     = 1;
    *selectKind   = showSelect;

  } else if (!strcmp (Tcl_DStringValue (&cmd), "UPDATE") ||
	     !strcmp (Tcl_DStringValue (&cmd), "INSERT") ||
	     !strcmp (Tcl_DStringValue (&cmd), "DELETE")) {
    *withResCount = 1;
  }

  Tcl_DStringFree (&cmd);
}

/* -------------------------------------------------------------- */

static void
getNullValue (interp, nullValue, specialNull)
     Tcl_Interp *interp;		/* Current interpreter. */
     char      **nullValue;		/* Representation of a NULL value. */
     char      **specialNull;		/* Representation of a SPECIAL NULL. */
{
  int dummyLength;			/* Out parameter of GetXxxFromObj. */

  *nullValue   = AdabasGetMsgString (interp, AdamsgNullvalue, &dummyLength);
  if (!nullValue)   *nullValue   = "";
  *specialNull = AdabasGetMsgString (interp, AdamsgSpecialnull, &dummyLength);
  if (!specialNull) *specialNull = "";
}

/*
 *----------------------------------------------------------------------
 *
 * AdabasCursorHandle --
 *
 *      Looks up the given cursorName in the cursor hash table.
 *
 * Results:
 *      The return value is normally a pointer to the CursorInfo structure
 *      of the found hash entry. If no cursor with the given name exists,
 *      NULL is returned.
 *
 * Side effects:
 *      The global array element adamsg(handle) will contain the cursorName.
 *
 *----------------------------------------------------------------------
 */

CursorInfo *
AdabasCursorHandle (interp, cursorName)
     Tcl_Interp  *interp;		/* Current interpreter. */
     char        *cursorName;		/* Name to look for. */
{
  Tcl_HashEntry *hashEntry;		/* Entry with cursor name and infoPrt. */
  CursorInfo    *cursorVars;		/* Found cursorHandle to return. */

  if (!(hashEntry = Tcl_FindHashEntry (&cursorHash, cursorName))) {
    AppendResult (interp, "No valid cursor handle: ");
    AppendResult (interp, cursorName);
    return (CursorInfo *) NULL;
  }
  cursorVars = (CursorInfo *) Tcl_GetHashValue (hashEntry);
  if (checkCursor (interp, cursorVars) == TCL_ERROR) {
    return (CursorInfo *) NULL;
  }

  AdabasSetMsgString (interp, AdamsgHandle, cursorName);
  return cursorVars;
}

/*
 *----------------------------------------------------------------------
 *
 * checkCursor --
 *
 *      Checks whether the given cursor is known and alive.
 *
 * Results:
 *      The return value is normally TCL_OK; if the cursor is invalid,
 *      TCL_ERRORNULL is returned and interp->result will point to the
 *      error message.
 *
 * Side effects:
 *      None
 *
 *----------------------------------------------------------------------
 */

static int
checkCursor (interp, cursor)
     Tcl_Interp *interp;		/* Current interpreter. */
     CursorInfo *cursor;		/* Cursor handle. */
{
  if (!cursor) {
    SetResult (interp, "Invalid cursorHandle (NULL)");
    return TCL_ERROR;
  }
  if (cursor->magic != CURSOR_MAGIC) {
    SetResult (interp, "Invalid cursorHandle (already dead)");
    return TCL_ERROR;
  }
  return TCL_OK;
}
